#include "ShareFileTransfer.h"
#include "ShareStrings.h"
#include "ShareWindow.h"
#include <app/Roster.h>
#include <kernel/OS.h>
#include <kernel/fs_attr.h>
#include <interface/ListView.h>
#include <interface/Region.h>
#include <storage/Mime.h>
#include <storage/NodeInfo.h>

#include "dataio/TCPSocketDataIO.h"
#include "iogateway/PortableMessageIOGateway.h"

#include "Colors.h"
#include "ShareUtils.h"
#include "md5.h"

#define MD5_DIGEST_SIZE 16
#define ANIM_RECT_SPACING 16
#define ANIM_RECT_WIDTH   16

const rgb_color PalePink = {255, 233, 233, 255};
const rgb_color PaleBlue = {233, 233, 255, 255};

const rgb_color BrightPink = {255, 200, 200, 255};
const rgb_color BrightBlue = {200, 200, 255, 255};

const rgb_color PolePink = {255, 216, 216, 255};
const rgb_color PoleBlue = {216, 216, 255, 255};

/* Abstract base class */
class MD5Looper : public BLooper 
{
public:
   MD5Looper()
   {
      // empty
   }

   status_t StartProcessing(int32 replyCode, const BDirectory & dir, PortableMessageRef msgRef, const BMessenger & replyTo, volatile bool * shutdownFlag)
   {
      _replyCode = replyCode;
      _dir       = dir;
      _msgRef    = msgRef;
      _replyTo   = replyTo; 
      _shutdownFlag = shutdownFlag;

      if (Run() >= 0)
      {
         BMessage whatever;
         PostMessage(&whatever);  // just to get us going
         return B_NO_ERROR;
      }
      else return B_ERROR;
   }

   virtual void MessageReceived(BMessage *) = 0;

   PortableMessageRef GetMessageRef() {return _msgRef;}

protected:
   void Reply()
   {
      BMessage reportDone(_replyCode);
      reportDone.AddPointer("from", this);
      _replyTo.SendMessage(&reportDone);
   }

   int32 _replyCode;
   BDirectory _dir;
   PortableMessageRef _msgRef;
   BMessenger _replyTo;
   volatile bool * _shutdownFlag;
};


/* Background thread used for reading MD5 hashcodes into our TRANSFER_COMMAND_FILE_LIST method, for sending */
class MD5SendLooper : public MD5Looper
{
public:
   virtual void MessageReceived(BMessage *)
   {
      PortableMessage & msg = *_msgRef.GetItemPointer();
      const char * nextStr;
      for (int i=0; (msg.FindString("files", i, &nextStr) == B_NO_ERROR); i++)
      {
         // Add current file length and md5, if any
         off_t fileOffset = 0;  // i.e. autodetect file size for offset
         uint8 digest[MD5_DIGEST_SIZE];
         BFile tempFile(&_dir, nextStr, B_READ_ONLY);
         (void) HashFileMD5(tempFile, fileOffset, digest, _shutdownFlag); // sets (fileOffset) on success
         msg.AddInt64("offsets", fileOffset);
         msg.AddData("md5", B_RAW_TYPE, digest, (fileOffset > 0) ? sizeof(digest) : 1);
      }
      Reply();
   }
};


/* Background thread used for reading MD5 hashcodes from files to verify them against the codes sent by the remote client */
class MD5ReceiveLooper : public MD5Looper
{
public:
   virtual void MessageReceived(BMessage *)
   {
      PortableMessage * msg = _msgRef.GetItemPointer();
      const char * nextFile;
      for (int i=0; (msg->FindString("files", i, &nextFile) == B_NO_ERROR); i++) 
      {
         // Check to see if the requestor already has the beginning of this file
         // If he does, we can do the old auto-resume-download trick!
         off_t offset = 0LL;
         if ((msg->FindInt64("offsets", i, &offset) == B_NO_ERROR)&&(offset > 0))
         {
            off_t uploadOffset = 0LL;  // default to starting the upload at the beginning of the file
            uint8 * hisDigest;
            size_t numBytes;
            if ((msg->FindData("md5", B_RAW_TYPE, i, (const void **) &hisDigest, &numBytes) == B_NO_ERROR)&&(numBytes == MD5_DIGEST_SIZE))
            {
               // Okay, the requestor has told us enough so that we can check our local file...
               uint8 myDigest[MD5_DIGEST_SIZE];
               BFile tempFile;
               if ((ShareFileTransfer::FindSharedFile(tempFile, _dir, nextFile, NULL) == B_NO_ERROR)&&
                   (HashFileMD5(tempFile, offset, myDigest, _shutdownFlag) == B_NO_ERROR)&&
                   (memcmp(hisDigest, myDigest, sizeof(myDigest)) == 0)) uploadOffset = offset;
            }
            msg->ReplaceInt64("offsets", i, uploadOffset);
         }
      }
      Reply();
   }
};

static uint32 MD5LooperPointerHashFunc(MD5Looper * const & ptr);
uint32 MD5LooperPointerHashFunc(MD5Looper * const & ptr)
{
   return (uint32) ptr;
}

ShareFileTransfer ::
ShareFileTransfer(const BDirectory & fileDir, const char * localSessionID) :
	_dir(fileDir),
	_fileSet(PortableStringHashFunc),
	_md5Loopers(MD5LooperPointerHashFunc),
	_currentFileStartTime(0),
	_currentFileStartOffset(0),
	_currentFileSize(-1),
	_currentFileBytesDone(-1),
	_mtt(NULL),
	_uploadSession(false),
	_onConnect(TRANSFER_COMMAND_CONNECTED_TO_PEER),
	_onDisconnect(TRANSFER_COMMAND_DISCONNECTED_FROM_PEER),
	_onEmpty(TRANSFER_OUTPUT_QUEUE_EMPTY),
	_bitmap(NULL),
	_localSessionID(localSessionID),
	_isAccepting(false),
	_isConnecting(false),
	_isConnected(false),
	_isWaitingOnLocal(true),
	_isWaitingOnRemote(false),
	_wasConnected(false),
	_errorOccurred(false),
	_isFinished(false),
	_currentFileIndex(0),
	_origFileSetSize(0),
	_animShift(0.0f),
	_lastTransferTime(0LL),
	_shutdownFlag(false)

{
   // empty
}

ShareFileTransfer ::
~ShareFileTransfer()
{
   // Clean up any MD5 loopers still in progress
   _shutdownFlag = true;  // get the loopers to stop immediately
   PortableHashtableIterator<MD5Looper *, bool> iter = _md5Loopers.GetIterator();
   MD5Looper * next;
   while(iter.GetNextKey(next)) if (next->Lock()) next->Quit();

   delete _mtt;
}

void 
ShareFileTransfer::
AddRequestedFileName(const char * fileName, off_t optStartByte)
{
   _fileSet.Put(PortableString(fileName), optStartByte);
}

status_t 
ShareFileTransfer::
InitConnectSession(const char * hostName, int port, const char * remoteSessionID)
{
   _remoteHostName = hostName;
   _remotePort = port;
   _remoteSessionID = remoteSessionID;
   _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(remoteSessionID);
   _fileSetIter = _fileSet.GetIterator();
   _currentFileBytesDone = _currentFileSize = -1;
   _currentFileStartOffset = 0;
   _currentFileStartTime = 0;
   _origFileSetSize = _fileSet.Size();
   _uploadSession = (_origFileSetSize == 0);

   BMessenger toMe(this);
   delete _mtt;  // just in case there is an old one around for some reason
   _mtt = new MessageTransceiverThread(toMe);
   if (_uploadSession) _mtt->SetOutgoingQueueDrainedMessage(&_onEmpty);  // so we get notified when it's time to send more data
   return B_NO_ERROR;
}

status_t
ShareFileTransfer ::
InitAcceptSession(const char * remoteSessionID)
{
   _isAccepting = true;
   _remoteSessionID = remoteSessionID;
   _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(remoteSessionID);
   _fileSetIter = _fileSet.GetIterator();
   _currentFileBytesDone = _currentFileSize = -1;
   _currentFileStartOffset = 0;
   _currentFileStartTime = 0;
   _origFileSetSize = _fileSet.Size();
   _uploadSession = (_origFileSetSize == 0);

   BMessenger toMe(this);
   delete _mtt;  // just in case there is an old one around for some reason
   _mtt = new MessageTransceiverThread(toMe);
   if (_uploadSession) _mtt->SetOutgoingQueueDrainedMessage(&_onEmpty);  // so we get notified when it's time to send more data

   // First try the "standard" ports (to satisfy those firewall-configging guys),
   // if none of those are available, we'll use any old port.
   status_t ret = B_ERROR;
   for (int i=DEFAULT_LISTEN_PORT; i<=DEFAULT_LISTEN_PORT+LISTEN_PORT_RANGE; i++)
   {
      if ((ret = _mtt->StartAcceptThread((i<DEFAULT_LISTEN_PORT+LISTEN_PORT_RANGE)?i:0, false, &_onConnect, &_onDisconnect)) == B_NO_ERROR) break;
   }

   if (ret == B_NO_ERROR) _isAccepting = true;
   UpdateTransferTime();  // enable inactivity timeout when waiting for a connection
   return ret;
}


status_t 
ShareFileTransfer::
InitSocketUploadSession(int socket, bool sendQueuedNotify)
{
   _isConnected = _uploadSession = true;
   _origFileSetSize = -1;  // we'll know this when we get the TRANSFER_COMMAND_FILE_LIST message

   BMessenger toMe(this);

   delete _mtt;  // just in case there is an old one around for some reason
   _mtt = new MessageTransceiverThread(toMe);
   _mtt->SetOutgoingQueueDrainedMessage(&_onEmpty);  // so we get notified when it's time to send more data
   status_t ret = _mtt->StartThread(new PortableMessageIOGateway(new TCPSocketDataIO(socket, false)), socket, &_onDisconnect);
   if ((ret == B_NO_ERROR)&&(sendQueuedNotify)) SendNotifyQueuedMessage();
   return ret;
}


void
ShareFileTransfer ::
SendNotifyQueuedMessage()
{
   PortableMessage * notifyMsg = _mtt->GetMessagePool()->GetObject();
   if (notifyMsg)
   {
      PortableMessageRef msgRef(notifyMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
      notifyMsg->what = TRANSFER_COMMAND_NOTIFY_QUEUED;
      MASSERT(msgRef.GetItemPointer(), "queuemsg:  no item pointer?");
      _mtt->SendOutgoingMessage(msgRef);
      ResetTransferTime(); // don't timeout if he's queued, in this case it's okay that the connection is idle
   }
}


void 
ShareFileTransfer::
MessageReceived(BMessage * msg)
{
   switch(msg->what)
   {
      case TRANSFER_COMMAND_MD5_RECV_READ_DONE:
      {
         MD5Looper * looper;
         if ((msg->FindPointer("from", (void **) &looper) == B_NO_ERROR)&&(_md5Loopers.Remove(looper)))
         {
            PortableMessageRef msgRef;
            msgRef = looper->GetMessageRef();
            if (looper->Lock()) looper->Quit();

            PortableMessage * pmsg = msgRef.GetItemPointer();
            const char * nextFile;
            for (int i=0; (pmsg->FindString("files", i, &nextFile) == B_NO_ERROR); i++)
            {
               uint64 offset = 0LL;
               (void) pmsg->FindInt64("offsets", i, (int64*) &offset);
               AddRequestedFileName(nextFile, offset);
            }

            int numFiles = _fileSet.Size();
            if (numFiles > 0) 
            {
               _origFileSetSize = numFiles;
               _fileSetIter = _fileSet.GetIterator();
               ((ShareWindow*)Looper())->RefreshTransferItem(this);
               DoUpload();
            }
            else AbortSession(true);   // he's gotta ask for something, otherwise why did he connect?
         }
      }
      break;

      case TRANSFER_COMMAND_MD5_SEND_READ_DONE:
      {
         MD5Looper * looper;
         if ((msg->FindPointer("from", (void **) &looper) == B_NO_ERROR)&&(_md5Loopers.Remove(looper)))
         {
            PortableMessageRef msgRef;
            msgRef = looper->GetMessageRef();
            if (looper->Lock()) looper->Quit();
            if (_mtt) _mtt->SendOutgoingMessage(msgRef);
            ((ShareWindow*)Looper())->RefreshTransferItem(this);
         }
      }
      break;

      case TRANSFER_COMMAND_CONNECTED_TO_PEER:
      {
         _isConnecting = false;
         _isConnected = _wasConnected = true;
         if (_mtt)  // paranoia
         {
            if (_uploadSession)
            {
               if (_isWaitingOnLocal) SendNotifyQueuedMessage();
            }
            else
            {
               // Tell the remote BeShare connection what files we want from him
               PortableMessage * fileRequest = _mtt->GetMessagePool()->GetObject();
               if (fileRequest)
               {
                  PortableMessageRef fileRequestRef(fileRequest, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
                  fileRequest->what = TRANSFER_COMMAND_FILE_LIST;
                  fileRequest->AddString("beshare:FromSession", _localSessionID);
                  PortableHashtableIterator<PortableString, off_t> iter = _fileSet.GetIterator();
                  PortableString * next;
                  while((next = iter.GetNextKey()) != NULL) fileRequest->AddString("files", next->Cstr());
                  
                  // Calculating the md5 hashes could take a long time if the files are big, so I'll do it in a background thread!
                  MD5Looper * md5Looper = new MD5SendLooper;
                  if (md5Looper->StartProcessing(TRANSFER_COMMAND_MD5_SEND_READ_DONE, _dir, fileRequestRef, BMessenger(this), &_shutdownFlag) == B_NO_ERROR) 
                  {
                     _md5Loopers.Put(md5Looper, true);
                     ((ShareWindow*)Looper())->RefreshTransferItem(this);
                  }
                  else 
                  {
                     delete md5Looper;  // ???
                     AbortSession(true);
                  }
               }
               else AbortSession(true);
            }
         }
         ((ShareWindow*)Looper())->FileTransferConnected(this);
      }
      break;

      case TRANSFER_COMMAND_DISCONNECTED_FROM_PEER:
      {
         _shutdownFlag = true;  // no sense reading any more md5 data if we're no longer connected!
         _isFinished = true;
         _isConnecting = _isConnected = _isAccepting = _isWaitingOnLocal = _isWaitingOnRemote = false;
         if ((_currentFileBytesDone < _currentFileSize)||(_currentFileIndex < _origFileSetSize)) _errorOccurred = true;
         ((ShareWindow*)Looper())->FileTransferDisconnected(this);
         delete _mtt;  // might as well get rid of it, it's useless now
         _mtt = NULL;
         ResetTransferTime();
      }
      break;

      case TRANSFER_OUTPUT_QUEUE_EMPTY:
         if (_uploadSession) DoUpload();
      break;

      case PORTABLE_MESSAGES_RECEIVED:
      {
         // This BMessage comes from our MessageTransceiverThread, and
         // tells us that there may be some incoming PortableMessages to look at.
         if (_mtt)
         {
            PortableMessageRef next;
            while((_mtt)&&(_mtt->GetNextIncomingMessage(next) == B_NO_ERROR)) PortableMessageReceived(next);
         }
      }
      break;            
   }
}


void
ShareFileTransfer ::
DoUpload()
{
   if (_isWaitingOnLocal) return;   // not yet!

   if (_mtt)
   {
      PortableMessage * msg = _mtt->GetMessagePool()->GetObject();
      if (msg)
      {
         UpdateTransferTime();  // avoid the moribund-connection-reaper
         PortableMessageRef msgRef(msg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
         if (_currentFileBytesDone >= 0)
         {
            ssize_t numBytes = _currentFile.Read(_scratchBuf, sizeof(_scratchBuf));
            if (numBytes > 0) 
            {
               msg->what = TRANSFER_COMMAND_FILE_DATA;
               msg->AddData("data", B_RAW_TYPE, _scratchBuf, numBytes);
               MASSERT(msgRef.GetItemPointer(), "datamsg:  no item pointer?");
               _mtt->SendOutgoingMessage(msgRef);
               _currentFileBytesDone += numBytes;
               ((ShareWindow *)Looper())->RefreshTransferItem(this);
               return;  
            }

            if (numBytes == 0)
            {
               // move to the next file
               _currentFileBytesDone = _currentFileSize = -1;
               _currentFileStartTime = 0;
               _currentFileStartOffset = 0;
               DoUpload();
               return;
            }
         }
         else
         {
            // Get the next file from our send list
            if (_fileSetIter.GetNextKey(_currentFileName))
            {
               _fileSetIter.GetNextValue(_currentFileBytesDone);  // get the starting byte to read from...
               _currentFileStartTime = system_time();
               _currentFileStartOffset = _currentFileBytesDone;

               if ((FindSharedFile(_currentFile, _dir, _currentFileName.Cstr(), &_currentFileEntry) == B_NO_ERROR)&&
                   (_currentFile.GetSize(&_currentFileSize) == B_NO_ERROR) &&
                   (_currentFile.Seek(_currentFileBytesDone, SEEK_SET) == _currentFileBytesDone))
               {
                  _currentFileIndex++;

                  if (((ShareWindow*)Looper())->LogUploads())
                  {
                     PortableString strng(str(STR_USER));
                     strng += _remoteSessionID;
                     if (_remoteUserName[0])
                     {
                        strng += str(STR_AKA);
                        strng += _remoteUserName;
                        strng += ")";
                     }
                     strng += str(STR_IS_DOWNLOADING);
                     strng += _currentFileName;
                     ((ShareWindow*)Looper())->LogMessage(LOG_INFORMATION_MESSAGE, strng.Cstr());
                  }

                  msg->AddString("beshare:File Name", _currentFileName.Cstr());
                  msg->AddInt64("beshare:File Size", _currentFileSize);
                  msg->AddString("beshare:FromSession", _localSessionID);
                  if (_currentFileBytesDone > 0LL) msg->AddInt64("beshare:StartOffset", _currentFileBytesDone);  // resume-download mode!

                  char attrName[B_ATTR_NAME_LENGTH];
                  BNodeInfo ni(&_currentFile);
                  if ((ni.InitCheck() == B_NO_ERROR)&&(ni.GetType(attrName) == B_NO_ERROR))
                  {
                     msg->AddString("beshare:Kind", attrName);
                     _bitmap = ((ShareWindow*)Looper())->GetBitmap(attrName);
                  }
                  else _bitmap = ((ShareWindow*)Looper())->GetBitmap(NULL); 

                  while(_currentFile.GetNextAttrName(attrName) == B_NO_ERROR)
                  {
                     struct attr_info attrInfo;
                     if ((_currentFile.GetAttrInfo(attrName, &attrInfo) == B_NO_ERROR)&&(attrInfo.size > 0))
                     {
                        char * attrData = new char[attrInfo.size];
                        ssize_t ret = _currentFile.ReadAttr(attrName, attrInfo.type, 0L, attrData, attrInfo.size);
                        if (ret == attrInfo.size) (void) msg->AddData(attrName, attrInfo.type, attrData, attrInfo.size);
                        delete [] attrData;
                     }
                  }

                  ((ShareWindow*)Looper())->RefreshTransferItem(this);

                  msg->what = TRANSFER_COMMAND_FILE_HEADER;
                  MASSERT(msgRef.GetItemPointer(), "headermsg: no itempointer");
                  _mtt->SendOutgoingMessage(msgRef);
                  DoUpload();
                  return;
               }
            }
            else
            {
               // All done!
               AbortSession(false);
               return;
            }
         }
      }
   }

   // If we got here, there was an error
   AbortSession(true);
}


status_t
ShareFileTransfer ::
FindSharedFile(BFile & setFile, const BDirectory & rootDir, const char * fileName, entry_ref * optSetEntryRef)
{
   // First try to find the file in our root dir
   if (setFile.SetTo(&rootDir, fileName, B_READ_ONLY) == B_NO_ERROR) 
   {
      // But if this entry is actually a directory, that doesn't count!
      BDirectory doubleCheck(&rootDir, fileName);
      if (doubleCheck.InitCheck() != B_NO_ERROR) 
      {
         if (optSetEntryRef) 
         {
            BEntry e;
            if (rootDir.FindEntry(fileName, &e, true) == B_NO_ERROR) (void) e.GetRef(optSetEntryRef);
         }
         return B_NO_ERROR;
      }
   }

   setFile.Unset();
   // If we didn't find it in this dir, check all our subdirs...
   BDirectory dir(rootDir);
   BEntry next;
   while(dir.GetNextEntry(&next, true) == B_NO_ERROR)
   {
      if (next.IsDirectory())
      {
         BDirectory subdir(&next);
         if ((subdir.InitCheck() == B_NO_ERROR)&&(FindSharedFile(setFile, subdir, fileName, optSetEntryRef) == B_NO_ERROR)) return B_NO_ERROR;
      }
   }

   // nope, can't find it anywhere
   return B_ERROR;  
}

void 
ShareFileTransfer ::
PortableMessageReceived(PortableMessageRef msgRef)
{
   PortableMessage * msg = msgRef.GetItemPointer();
   UpdateTransferTime();
   switch(msg->what)
   {
      case TRANSFER_COMMAND_NOTIFY_QUEUED:
         if (_uploadSession == false) 
         {
            _isWaitingOnRemote = true;
            ((ShareWindow *)Looper())->RefreshTransferItem(this);
            ResetTransferTime(); // If we're queued, don't timeout, as we may not be getting data for a long time anyway
         }
      break;

      // Received when the remote peer wishes to tell us what files he wants to download from us
      case TRANSFER_COMMAND_FILE_LIST:
      {
         const char * check;  // not used here
         if ((_uploadSession)&&(msg->FindString("files", &check) == B_NO_ERROR))  // make sure he's asking for at least one file
         {
            PortableString old = _remoteSessionID;

            if (msg->FindString("beshare:FromSession", _remoteSessionID) == B_NO_ERROR) 
            {
               _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(_remoteSessionID.Cstr());
            }
            else _remoteUserName = str(STR_UNKNOWN);

            // in case our session ID allows us to start...
            if (old != _remoteSessionID) ((ShareWindow*)Looper())->DequeueTransferSessions();

            MD5Looper * md5Looper = new MD5ReceiveLooper();
            if (md5Looper->StartProcessing(TRANSFER_COMMAND_MD5_RECV_READ_DONE, _dir, msgRef, BMessenger(this), &_shutdownFlag) == B_NO_ERROR) 
            {
               _md5Loopers.Put(md5Looper, true);
               ((ShareWindow*)Looper())->RefreshTransferItem(this);
            }
            else 
            {
               delete md5Looper;
               AbortSession(true);
            }
         }
         else AbortSession(true);  // you're going the wrong way buddy!
      }
      break;      

      // When downloading, this indicates the start of the next incoming file
      case TRANSFER_COMMAND_FILE_HEADER:
      {
         PortableString old(_remoteSessionID);
         if ((_uploadSession == false)                                             &&
             (msg->FindString("beshare:File Name", _currentFileName) == B_NO_ERROR)&&
             (msg->FindInt64("beshare:File Size", &_currentFileSize) == B_NO_ERROR)&&
             (msg->FindString("beshare:FromSession", _remoteSessionID) == B_NO_ERROR)&&
             (_fileSet.Remove(_currentFileName)))  // note that while downloading we ignore the value of this hashtable entry
         {
            _currentFileStartTime = system_time();
            off_t startByte = 0LL;

            bool append = false;
            if ((msg->FindInt64("beshare:StartOffset", &startByte) == B_NO_ERROR)&&(startByte > 0LL))
            {
               // Double check:  Do we have a file named _currentFileName that is exactly the right size?
               // If so, go to append mode; if not, fall back to the old way
               off_t curFileSize;
               if ((_currentFile.SetTo(&_dir, _currentFileName.Cstr(), B_READ_ONLY) == B_NO_ERROR)&&
                   (_currentFile.GetSize(&curFileSize) == B_NO_ERROR)&&(curFileSize == startByte)) append = true;
            }

            if (((append)||(GenerateNewFilename(_currentFileName) == B_NO_ERROR))&&
                (_currentFile.SetTo(&_dir, _currentFileName.Cstr(), B_WRITE_ONLY|B_CREATE_FILE|(append?B_OPEN_AT_END:0)) == B_NO_ERROR))
            {
               BEntry e;
               if (_dir.FindEntry(_currentFileName.Cstr(), &e, true) == B_NO_ERROR) (void) e.GetRef(&_currentFileEntry);
               _isWaitingOnRemote = false;  // if we're getting data, we must not be waiting anymore!
               _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(_remoteSessionID.Cstr());
               _currentFileIndex++;
               _currentFileStartOffset = _currentFileBytesDone = startByte;
               PortableMessageFieldNameIterator iter = msg->GetFieldNameIterator();
               PortableString fieldName;
               while(iter.GetNextFieldName(fieldName))
               {
                  if (strncmp(fieldName.Cstr(), "beshare:", 8))  // don't save our attributes, they aren't necessary
                  {
                     const void * attrData;
                     size_t attrSize;
                     int32 c;
                     type_code type;
                     if ((msg->GetInfo(fieldName.Cstr(), &type, &c)                   == B_NO_ERROR)&&
                         (msg->FindData(fieldName.Cstr(), type, &attrData, &attrSize) == B_NO_ERROR))
                     {
                        (void)_currentFile.WriteAttr(fieldName.Cstr(), type, 0, attrData, attrSize);
                     }
                  }
               }

               BNodeInfo ni(&_currentFile);
               const char * mimeString;
               if ((ni.InitCheck() == B_NO_ERROR)&&(msg->FindString("beshare:Kind", &mimeString) == B_NO_ERROR))
               {
                  ni.SetType(mimeString);
                  _bitmap = ((ShareWindow*)Looper())->GetBitmap(mimeString);
               }
               else _bitmap = ((ShareWindow*)Looper())->GetBitmap(NULL);

               ((ShareWindow*)Looper())->RefreshTransferItem(this);

               // in case our name allows us to start...
               if (old != _remoteSessionID) ((ShareWindow*)Looper())->DequeueTransferSessions();
            }
            else AbortSession(true);
         }
         else AbortSession(true);
      }
      break;

      case TRANSFER_COMMAND_FILE_DATA:
      {
         const uint8 * data;
         size_t numBytes;
         if ((_uploadSession == false)&&(msg->FindData("data", B_RAW_TYPE, (const void **) &data, &numBytes) == B_NO_ERROR))
         {
            _isWaitingOnRemote = false;  // obviously....
            if (_currentFile.Write(data, numBytes) == (ssize_t) numBytes) 
            {
               _currentFileBytesDone += numBytes;
               ((ShareWindow*)Looper())->RefreshTransferItem(this);
            }
            else 
            {
               PortableString errStr(str(STR_ERROR_WRITING_TO_FILE));
               errStr += _currentFileName;
               ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
               AbortSession(true);
            }
         }
         else AbortSession(true);
      }
      break;
   }
}

status_t
ShareFileTransfer ::
GenerateNewFilename(PortableString & str) const
{
   for (int i=0; i<50; i++)
   {
      BFile f(&_dir, str.Cstr(), B_READ_ONLY);
      if (f.InitCheck() != B_NO_ERROR) return B_NO_ERROR;

      int lastSpaceIdx = str.LastIndexOf(' ');
      if ((lastSpaceIdx >= 0)&&(lastSpaceIdx < (int) str.Length()-1))
      {
         const char * cs = str.Cstr()+lastSpaceIdx+1;
         if ((cs[0] >= '0')&&(cs[0] <= '9'))
         {
            char buf[20];
            sprintf(buf, "%i", atoi(cs)+1);
            str = str.Substring(0, lastSpaceIdx+1) + buf;
            continue;
         }
      }
      str += " 2";  // default permutation
   }
   return B_ERROR;  // give up, something has gone horribly wrong
}


void
ShareFileTransfer::
AbortSession(bool error)
{
   if (_mtt)
   {
      ((ShareWindow*)Looper())->FileTransferDisconnected(this);
      if (error == false) _mtt->WaitForAllMessagesToBeSent();
      delete _mtt;
      _mtt = NULL;
   }
   _isConnecting = _isAccepting = _isConnected = _isWaitingOnLocal = _isWaitingOnRemote = false;
   _currentFileBytesDone = _currentFileSize = -1;
   _currentFileStartOffset = 0;
   _errorOccurred = error;
   _isFinished = true;
   ResetTransferTime();
}


void 
ShareFileTransfer ::
Update(BView * lv, const BFont * font)
{
   BListItem::Update(lv, font);
   SetHeight(TRANSFER_ITEM_HEIGHT);
   font_height fontAttrs;
   font->GetHeight(&fontAttrs);
   _fontHeight = ceil(fontAttrs.ascent) + ceil(fontAttrs.descent);
}

void 
ShareFileTransfer ::
DrawItem(BView * lv, BRect itemRect, bool /*complete*/)
{
   bool active = ((_isWaitingOnLocal)||(_isWaitingOnRemote)||(_isConnected)||(_isConnecting));
   rgb_color backgroundColor = active ? (_uploadSession ? PalePink : PaleBlue) : Grey;
 
   BRegion Region;
   Region.Include(itemRect);
   lv->ConstrainClippingRegion(&Region);

   BRect colorRect = itemRect;
   colorRect.InsetBy(1.0f, 1.0f);  // avoid flicker by not drawing on the outer layer of pixels 
   PortableString text;
   float percent = (_currentFileSize > 0) ? ((float)_currentFileBytesDone)/_currentFileSize : 0.0f;
   BRect fillRectRight(colorRect);
   if (active) fillRectRight.left = (colorRect.left + colorRect.Width()*percent);
   lv->SetHighColor(backgroundColor);
   lv->FillRect(fillRectRight);

   if (active)
   {
      const rgb_color barColor = _uploadSession ? BrightPink : BrightBlue;
      BRect fillRectLeft(colorRect);
      fillRectLeft.right = fillRectRight.left;
      lv->SetHighColor(barColor);
      lv->FillRect(fillRectLeft);
      if (percent > 0.5f) backgroundColor = barColor;  // do text background from whoever has majority

      // Do little scrolly barber pole animation thingy
      {
         _animShift += 2.0f;
         if (_animShift > ANIM_RECT_SPACING*2.0f) _animShift -= (ANIM_RECT_SPACING*2.0f);

         lv->SetHighColor(_uploadSession ? PolePink : PoleBlue);

         BRegion barberRegion;
         barberRegion.Include(fillRectLeft);
         lv->ConstrainClippingRegion(&barberRegion);
         for (float rl=fillRectRight.right+(ANIM_RECT_SPACING*4)-_animShift; rl >= -ANIM_RECT_SPACING; rl -= ANIM_RECT_SPACING*2.0f)
         {
            const BPoint points[4] = {
               BPoint(rl,                   fillRectLeft.top),
               BPoint(rl-ANIM_RECT_WIDTH,   fillRectLeft.top),
               BPoint(rl-ANIM_RECT_SPACING*4, fillRectLeft.bottom),
               BPoint(ANIM_RECT_WIDTH+(rl-ANIM_RECT_SPACING*4), fillRectLeft.bottom)
            };
            lv->FillPolygon(points, ARRAYITEMS(points));
         }
         lv->ConstrainClippingRegion(&Region);
      }
   }

   if (_errorOccurred) text = str(STR_AN_ERROR_OCCURRED);
   else
   {
           if (_isWaitingOnLocal) text = str(STR_QUEUED_LOCAL_MACHINE_TOO_BUSY);
      else if (_isWaitingOnRemote) text = str(STR_QUEUED_REMOTE_MACHINE_TOO_BUSY);
      else if (_isConnected)
      {
         if (_md5Loopers.Size() > 0) text = str(STR_EXAMINING_FILES);
         else
         {
            text = _uploadSession ? str(STR_SENT) : str(STR_RCVD);

            if (_currentFileSize > 0)
            {
               char doneBuf[64], sizeBuf[64];
               GetByteSizeString(_currentFileBytesDone, doneBuf);
               GetByteSizeString(_currentFileSize, sizeBuf);
               char buf[100];
               sprintf(buf, "%s/%s", doneBuf, sizeBuf);
               text += buf;
            }

            if (_currentFileStartTime > 0)
            {
               bigtime_t elapsedTime = system_time() - _currentFileStartTime;
               if (elapsedTime > 0)
               {
                  // Add performance info, e.g. (43KB/sec, 5:05)
                  text += " (";
     
                  uint64 bytesPerSecond = (((_currentFileBytesDone-_currentFileStartOffset)*1000000LL)/elapsedTime);
                  uint64 bytesToGo      = _currentFileSize-_currentFileBytesDone;
                  uint64 secondsToGo    = (bytesPerSecond > 0) ? (bytesToGo/bytesPerSecond) : 0;

                  char buf[64];
                  GetByteSizeString(bytesPerSecond, buf); 
                  text += buf;
                  text += str(STR_SEC);
              
                  if (secondsToGo >= (60*60)) sprintf(buf, "%Lu:%02Lu:%02Lu", secondsToGo/(60*60), (secondsToGo/60)%60, secondsToGo%60);
                                         else sprintf(buf, "%Lu:%02Lu", secondsToGo/60, secondsToGo%60);
                  text += buf;
                  text += ")";
               }
            }
         }
      }
      else
      {
              if (_isAccepting) text = str(STR_AWAITING_CALLBACK);
         else if (_isConnecting) text = str(STR_CONNECTING);
         else if (_isFinished) 
         {
            if (_wasConnected) text = (_fileSet.Size() == 0) ? str(STR_DOWNLOAD_COMPLETE) : str(STR_DOWNLOAD_ABORTED);
                          else text = str(STR_COULDNT_CONNECT);
         }
         else text = _uploadSession ? str(STR_PREPARING_TO_UPLOAD) : str(STR_PREPARING_TO_DOWNLOAD);
      }
   }

   lv->SetLowColor(backgroundColor);
   lv->SetHighColor(Black);

   char line2[64];
   sprintf(line2, "(%i/%i) %s ", _currentFileIndex, _origFileSetSize, _uploadSession?str(STR_TO):str(STR_FROM));
   PortableString line2Str(line2);
   line2Str += _remoteUserName;
   lv->DrawString(line2Str.Cstr(), BPoint(2.0, colorRect.bottom-(_fontHeight+4.0f)));

   lv->DrawString(_currentFileName.Cstr(), BPoint((_bitmap?_bitmap->Bounds().Width()+8.0f:4.0f), colorRect.top+_fontHeight));
   lv->DrawString(text.Cstr(), BPoint(colorRect.left+2.0f,colorRect.bottom-3.0f));
   if (_bitmap)
   {
      lv->SetDrawingMode(B_OP_OVER);
      lv->DrawBitmap(_bitmap, BPoint(colorRect.left + 2.0f, colorRect.top));
   }

   // Do border highlighting
   bool selected = IsSelected();
   lv->SetHighColor(selected ? Black : White);
   lv->StrokeLine(BPoint(itemRect.left, itemRect.top), BPoint(itemRect.right, itemRect.top));
   lv->StrokeLine(BPoint(itemRect.left, itemRect.top), BPoint(itemRect.left, itemRect.bottom));
   lv->SetHighColor(selected ? White : Black);
   lv->StrokeLine(BPoint(itemRect.left, itemRect.bottom), BPoint(itemRect.right, itemRect.bottom));
   lv->StrokeLine(BPoint(itemRect.right, itemRect.top), BPoint(itemRect.right, itemRect.bottom));

   lv->ConstrainClippingRegion(NULL);
}



void
ShareFileTransfer ::
BeginTransfer()
{
   if (_isWaitingOnLocal)
   {
      bool startConnect = false;

      _isWaitingOnLocal = false;
      if (_uploadSession)
      {
         if (_isConnected) 
         {
            if (_fileSet.Size() > 0) DoUpload();
         }
         else startConnect = true;
      }
      else
      {
         if (_isAccepting)
         {
            int port = _mtt ? _mtt->GetPort() : -1;
            if (port >= 0) ((ShareWindow*)Looper())->SendConnectBackRequestMessage(_remoteSessionID.Cstr(), port);
            else ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, str(STR_ERROR_STARTING_DOWNLOAD_NO_ACCEPT_PORT));
         }
         else startConnect = true;
      }

      if (startConnect)
      {
         if (_mtt)
         {
            status_t ret = _mtt->StartConnectThread(_remoteHostName.Cstr(), _remotePort, &_onConnect, &_onDisconnect);
            if (ret == B_NO_ERROR) _isConnecting = true;
                              else ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, str(STR_ERROR_STARTING_DELAYED_CONNECT));
         }
         else ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, str(STR_ERROR_STARTING_DELAYED_CONNECT_NO_TRANSCEIVER_THREAD));
      }

      ((ShareWindow *)Looper())->RefreshTransferItem(this);
   }
}

status_t
ShareFileTransfer ::
LaunchCurrentItem()
{
    return be_roster->Launch(&_currentFileEntry);
}

void
ShareFileTransfer ::
UpdateTransferTime()
{
   _lastTransferTime = ((_isWaitingOnLocal)||(_isWaitingOnRemote)) ? 0LL : system_time();
}

void
ShareFileTransfer ::
ResetTransferTime()
{
   _lastTransferTime = 0LL;
}
