#ifndef SHARE_WINDOW_H
#define SHARE_WINDOW_H

#include <app/Message.h>
#include <app/MessageFilter.h>
#include <interface/Window.h>
#include <interface/TextControl.h>
#include <interface/TextView.h>
#include <interface/Menu.h>
#include <storage/Directory.h>

#include "queue/PortableQueue.h"
#include "message/PortableMessage.h"
#include "besupport/MessageTransceiverThread.h"

#include "PrefilledBitmap.h"
#include "CLVListItem.h"

#include "ShareConstants.h"

class SplitPane;
class ShareNetClient;
class ShareFileTransfer;
class RemoteUserItem;
class RemoteFileItem;
class ShareColumn;
class ShareMIMEInfo;
class ColumnListView;

class ShareWindow : public BWindow, public BMessageFilter
{
public:
   ShareWindow();
   ~ShareWindow();

   virtual void MessageReceived(BMessage * msg);

   virtual bool QuitRequested();

   virtual filter_result Filter(BMessage *msg, BHandler ** target);

   void SetConnectStatus(bool isConnecting, bool isConnected);
   void LogMessage(LogMessageType type, const char * text, const char * optSessionID=NULL, const rgb_color * optTextColor = NULL, bool isPersonal = false);
   void LogHelp(const char * cmd, int tokenID, int descID);

   void PutUser(const char * sessionID, const char * userName, const char * hostName, int userPort);
   void RemoveUser(const char * sessionID);
   
   void PutResult(const char * sessionID, const char * fileName, bool isFirewalled, PortableMessageRef fileInfo);
   void RemoveResult(const char * sessionID, const char * fileName);

   void FileTransferConnected(ShareFileTransfer * who);
   void FileTransferDisconnected(ShareFileTransfer * who);

   // Called by a ShareFileTransfer who wants his GUI updated.
   void RefreshTransferItem(ShareFileTransfer * item);

   // Returns a bitmap representing the icon of the given type, if any
   const BBitmap * GetBitmap(const char * mimeType);
 
   // Given a session ID, returns the user name for that ID.
   // If the session ID isn't known, "<unknown>" is returned.
   const char * GetUserNameForSession(const char * sessionID) const;

   // When doing a lot of add/remove file items, it's best to
   // bracket your calls with these for, efficieny in updates the GUI.
   void BeginBatchFileResultUpdate();
   void EndBatchFileResultUpdate();

   // Called when someone wants us to connect to him because we're behind a firewall
   // and that's the only way to download files from us.
   void ConnectBackRequestReceived(const char * targetSessionID, int port);

   void SetUserBandwidth(const char * sessionID, const char * label, uint32 bps);
   void SetUserFileCount(const char * sessionID, int32 fileCount);
   void SetUserIsFirewalled(const char * sessionID, bool firewalled); 

   // Tells the net client to send a message to another BeShare client, asking it
   // to connect back to us.
   void SendConnectBackRequestMessage(const char * remoteSessionID, int port);

   // Returns true iff we want uploads logged to the chat view
   bool LogUploads() const;

   // Returns true iff the user wants to see messages about usernames, status, etc.
   bool LogUserEvents() const;

   // See if any more transfers can be started
   void DequeueTransferSessions();

   // Returns true iff we're firewalled
   bool GetFirewalled() const;

private:
   static int CompareFunc(const CLVListItem* item1, const CLVListItem* item2, int32 sort_key);
   static int UserCompareFunc(const CLVListItem* item1, const CLVListItem* item2, int32 sort_key);
   int Compare(const RemoteFileItem * rf1, const RemoteFileItem * rf2, int32 sort_key) const;

   friend class RemoteUserItem; 
   friend class RemoteFileItem;
   friend class ShareColumn;
   friend class ShareMIMEInfo;

   // called by RemoteUserItem
   void AddFileItem(RemoteFileItem * item);
   void RemoveFileItem(RemoteFileItem * item);
   void RefreshFileItem(RemoteFileItem * item);
   void RefreshUserItem(RemoteUserItem * item);

   const char * GetFileCellText(const RemoteFileItem * item, int32 columnIndex) const;
   const BBitmap * GetBitmap(const RemoteFileItem * file, int32 columnIndex) const;

   // 'what' codes that are used internally by this class
   enum 
   {
      SHAREWINDOW_COMMAND_RECONNECT_TO_SERVER = '$win',
      SHAREWINDOW_COMMAND_DISCONNECT_FROM_SERVER,
      SHAREWINDOW_COMMAND_ABOUT,
      SHAREWINDOW_COMMAND_SEND_CHAT_TEXT,
      SHAREWINDOW_COMMAND_CHANGE_FILE_NAME_QUERY,
      SHAREWINDOW_COMMAND_USER_CHANGED_NAME,
      SHAREWINDOW_COMMAND_USER_CHANGED_SERVER,
      SHAREWINDOW_COMMAND_ENABLE_QUERY,
      SHAREWINDOW_COMMAND_DISABLE_QUERY,
      SHAREWINDOW_COMMAND_TOGGLE_COLUMN,
      SHAREWINDOW_COMMAND_BEGIN_DOWNLOADS,
      SHAREWINDOW_COMMAND_CANCEL_DOWNLOADS,
      SHAREWINDOW_COMMAND_CLEAR_FINISHED_DOWNLOADS,
      SHAREWINDOW_COMMAND_RESULT_SELECTION_CHANGED,
      SHAREWINDOW_COMMAND_REMOVE_SESSION,
      SHAREWINDOW_COMMAND_NEW_PEER_SESSION,
      SHAREWINDOW_COMMAND_TOGGLE_FIREWALLED,
      SHAREWINDOW_COMMAND_TOGGLE_LOG_UPLOADS,
      SHAREWINDOW_COMMAND_CLEAR_CHAT_LOG,
      SHAREWINDOW_COMMAND_SET_UPLOAD_LIMIT,
      SHAREWINDOW_COMMAND_SET_ADVERTISED_BANDWIDTH,
      SHAREWINDOW_COMMAND_RESET_LAYOUT,
      SHAREWINDOW_COMMAND_SET_DOWNLOAD_LIMIT,
      SHAREWINDOW_COMMAND_SELECT_USER,
      SHAREWINDOW_COMMAND_OPEN_SHARED_FOLDER,
      SHAREWINDOW_COMMAND_OPEN_DOWNLOADS_FOLDER,
      SHAREWINDOW_COMMAND_TOGGLE_LOG_USER_EVENTS,
      SHAREWINDOW_COMMAND_TOGGLE_LOG_QUERY_CHANGES,
      SHAREWINDOW_COMMAND_TOGGLE_FULL_USER_QUERIES,
      SHAREWINDOW_COMMAND_USER_SELECTED_SERVER,
      SHAREWINDOW_COMMAND_SELECT_LANGUAGE,
      SHAREWINDOW_COMMAND_LAUNCH_TRANSFER_ITEM,
      SHAREWINDOW_COMMAND_TOGGLE_LOG_CHAT,
      SHAREWINDOW_COMMAND_CHECK_FOR_MORIBUND_CONNECTIONS,
      SHAREWINDOW_COMMAND_TOGGLE_BEEP_WHEN_MENTIONED,
      SHAREWINDOW_COMMAND_SET_UPLOAD_PER_USER_LIMIT,
      SHAREWINDOW_COMMAND_TOGGLE_LOG_TIMESTAMPS,
      SHAREWINDOW_COMMAND_TOGGLE_AUTOCLEAR_COMPLETED_DOWNLOADS,
SHAREWINDOW_COMMAND_CHECK_ALICE
   };

   status_t GetAppSubdir(const char * subDirName, BDirectory & setSubDirRef, bool createIfNecessary) const;
   void UpdateConnectStatus(bool updateTitleToo);
   void UpdateQueryEnabledStatus();
   void SetQueryEnabled(bool enabled, const char * optLogTitle, bool putInQueryMenu = true);

   void ClearUsers();   // also clears results, as results are held by users
   void ClearResults();

   BMenu * MakeLimitSubmenu(const BMessage & msg, uint32 code, const char * label, const char * fieldName, uint32 & var);

   void AddBandwidthOption(BMenu * qMenu, const char * label, int bps);

   void CreateColumn(ShareMIMEInfo * optMIMEInfo, const char * columnName, bool remote, char shortcutKey);
   ShareMIMEInfo * CacheMIMETypeInfo(const char * mimeString);

   void UpdateDownloadButtonStatus();
   void RequestDownloads();

   void OpenTrackerFolder(const BDirectory & dir);

   BDirectory _shareDir;
   BDirectory _downloadsDir;

   bool _queryEnabled;

   bool _isConnecting;
   bool _isConnected;
   PortableString _connectedTo;

   BButton * _enableQueryButton;
   BButton * _disableQueryButton;

   BButton * _clearFinishedDownloadsButton;
   BButton * _requestDownloadsButton;
   BButton * _cancelTransfersButton;

   BMenu        * _queryMenu;

   BMenu        * _serverMenu;
   BTextControl * _serverEntry;

   BMenu        * _userNameMenu;
   BTextControl * _userNameEntry;

   BTextControl * _fileNameQueryEntry;

   BTextView * _chatText;
   BScrollView * _chatScrollView;
   BTextControl * _textEntry;
   BMenuItem * _clearChatItem;

   BMenuBar * _menuBar;

   BMenuItem * _firewalled;
   BMenuItem * _logUploads;
   BMenuItem * _logUserEvents;
   BMenuItem * _logQueryChanges;
   BMenuItem * _logChat;
   BMenuItem * _logTimestamps;
   BMenuItem * _connectMenuItem;
   BMenuItem * _disconnectMenuItem;
   BMenuItem * _fullUserQueries;
   BMenuItem * _beepWhenMentioned;
   BMenuItem * _autoClearCompletedDownloads;
 
   BMenu * _attribMenu;

   ColumnListView * _resultsView;
   ColumnListView * _usersView;

   int64 _bytesShown;

   BListView * _transferList;

   ShareNetClient * _netClient;

   PrefilledBitmap _defaultBitmap;

   PortableHashtable<const char *, RemoteUserItem *> _users;

   PortableHashtable<PortableString, float> _activeAttribs;        /* attribs the user wants to see, and their widths */
   PortableHashtable<const char *, ShareColumn *> _columns;        /* columns that are available */
   PortableHashtable<const char *, BMenuItem *> _attribMenuItems;  /* toggles for available columns */
   PortableHashtable<const char *, ShareMIMEInfo *> _mimeInfos;    /* cached MIME information */

   PortableHashtable<ShareMIMEInfo *, bool> _emptyMimeInfos;       /* MIME infos that aren't in the BMenu yet */

   PortableQueue<PortableString> _chatHistory;  // last <n> Items that you previously typed in
   int _chatHistoryPosition;                    // -1, or index into (_chatHistory)

   MessageTransceiverThread * _shareServer;  // used to accept incoming connections from peers

   SplitPane * _resultsTransferSplit;
   SplitPane * _mainSplit;
   SplitPane * _chatUsersSplit;

   BList _tempAddList;    // batch up items to be added, for efficiency

   uint32 CountActiveSessions(bool upload, const char * optForSession) const;

   uint32 _maxSimultaneousUploadSessions;    // the max # we allow (to avoid too much congestion)
   uint32 _maxSimultaneousUploadSessionsPerUser;  // the max # we allow per user (to avoid too much congestion)
   uint32 _maxSimultaneousDownloadSessions;  // the max # we allow (to avoid too much congestion)
   uint32 _uploadBandwidth;

   BMessageRunner * _connectionReaper;   // Every so often, causes us to check for moribund connections & kill them

   void RestoreSplitPane(SplitPane * sp, const BMessage & settingsMsg, const char * name) const;
   void SaveSplitPane(const SplitPane * sp, BMessage & settingsMsg, const char * name) const;

   void ResetLayout();

   // If a session is queued and waiting, this method will start it up.
   void DequeueTransferSessions(bool upload);

   void UpdateTitleBar();
   void AddServerItem(const char * serverName);

   int32 _language;

BMessageRunner * _aliceRunner;
PortableString _alice;
};

#endif
